@@ -35,28 +35,28 @@ from __future__ import print_function
  minimization of a function over a box in R^n
 
  Input:
- file          name of file for input and output
+  file         name of file for input and output
                if nargin < 5, the program continues a previous run and
                reads from file.mat the output is (again) stored in file.mat
 
 ^^do not use file - store variables globally,
 or make them available to be passed in?
 
- x             the rows are a set of new points entering the
+  x            the rows are a set of new points entering the
                optimization algorithm together with their function
                values
- f             matrix containing the corresponding function values
+  f            matrix containing the corresponding function values
                and their uncertainties, i.e., f(j,1) = f(x(j))
                and f(j,2) = df(x(j))
                a value f(j,2)<=0 indicates that the corresponding
                uncertainty is not known, and the program resets it to
                sqrt(numpy.spacing(1))
- config        structure variable defining the box [u,v] in which the
+  config       structure variable defining the box [u,v] in which the
                points are to be generated, the number nreq of
                points to be generated and the probability p that a
                point of type 4 is generated
                config = struct('bounds',{u,v},'nreq',nreq,'p',p)
- dx            only used for the definition of a new problem (when
+  dx           only used for the definition of a new problem (when
                the program should continue from the values stored in
                file.mat, the call should have only 4 input parameters!)
                n-vector (n = dimension of the problem) of minimal
@@ -65,37 +65,41 @@ or make them available to be passed in?
                coordinate i
 
  Output:
- request       nreq x (n+3)-matrix
-               request(j,1:n) is the jth newly generated point,
-               request(j,n+1) is its estimated function value and
-               request(j,n+3) indicates for which reason the point
-               request(j,1:n) has been generated
-               request(j,n+3) = 1 best prediction
+  request      nreq x (n+3)-matrix
+               request[j,1:n] is the jth newly generated point,
+               request[j,n+1] is its estimated function value and
+               request[j,n+3] indicates for which reason the point
+               request[j,1:n] has been generated
+               request[j,n+3] = 1 best prediction
                               = 2 putative local minimizer
                               = 3 alternative good point
                               = 4 explore empty region
                               = 5 to fill up the required number of
                               function values if too little points of
                               the other classes are found
- xbest         current best point
- fbest         current best function value (i.e. function value at xbest)
+  xbest        current best point
+  fbest        current best function value (i.e. function value at xbest)
 """
 
-from ._common import Result, Stats
-from ._gen_utils import diag, rsort, max_, min_, find, extend, rand, sort
+from SQCommon import Result, ObjectiveFunction
+from ._gen_utils import diag, max_, min_, find, extend, rand, sort
+from ._optset    import optset
 from ._snobinput import snobinput
 from ._snoblocf  import snoblocf, snobround
 from ._snoblp    import snoblp
 from ._snobnan   import snobnan
 from ._snobpoint import snobpoint
 from ._snobqfit  import snobqfit
+from ._snobsplit import snobsplit
 from ._snobupdt  import snobupdt
 from ._snob5     import snob5
-import logging, math, numpy
+import logging
+import math
+import numpy
 
 __all__ = ['minimize', 'log']
 
-log = logging.getLogger('SQSnobFit')
+log = logging.getLogger('SKQ.SnobFit')
 
 
 # dummy save/load (can always pickle but seems superflous; deal with this
@@ -139,77 +143,109 @@ def fill_request(request, func, nparams):
     for i in range(len(request)):
         x[i] = request[i][0:nparams]
         try:
-            f[i] = (func(x[i]), math.sqrt(numpy.spacing(1)))
+            res = func(x[i])
+            err = math.sqrt(numpy.spacing(1))
+            if type(res) == tuple:
+                if res[1] != 0:
+                    err = res[1]
+                res = res[0]
+            f[i] = (res, err)
         except Exception as e:
-            print('Function evaluation failed:', e)
+            log.Error('Function evaluation failed: %s', str(e))
             f[i] = numpy.nan
     return x, f
 
 
 #-----
-def minimize(f, x0, bounds, budget, optin=None, **optkwds):
-    # The user-facing API is the equivalent of snobdriver, providing the loop
-    # over the "internal" snobfit function.
+def minimize(f, x0, bounds, budget, optin={}, **optkwds):
+  # The user-facing API is the equivalent of snobdriver, providing the loop
+  # over the "internal" snobfit function.
     if budget <= 0:
         budget = 100000
 
+    if type(x0) != numpy.ndarray:
+        x0 = numpy.array(x0)
+
+    if type(bounds) != numpy.ndarray:
+        bounds = numpy.array(bounds)
+
+    objfunc = ObjectiveFunction(f, options={'simple_function' : True})
+
     minfcall = 10;      # minimum number of function values before
                         # considering stopping
-    nstop = 10           # number of times no improvement is tolerated
 
-    # calculate resolution vector from the bounds
+  # calculate resolution vector from the bounds
     dx = (bounds[:,1]-bounds[:,0])*1E-5
 
-    # setup parameters (TODO: use optin/optkwds)
-    config = {"bounds": bounds, "nreq": len(x0)+6, "p": .5}
+  # setup parameters (TODO: use optin/optkwds)
+    if type(optin) == dict:
+        options = optset(**dict(optin, **optkwds))
+    else:
+        options = optset(optin, **optkwds)
+
+    config = {"bounds": bounds, "nreq": 2*len(bounds)+6, "p": .5}
+    if optin is not None:
+        if options.maxmp is not None:
+            config["nreq"] = options.maxmp
+
+    nstop = options.maxfail      # number of times no improvement is tolerated
+
     nparams = len(bounds)
 
-    # initial call with empty list of input points (establishes initial request)
-    request, xbest, fbest = snobfit(
-        numpy.array([]).reshape(0, len(bounds)), numpy.array([]).reshape(0,2), config, dx)
-    if optin and 'verbose' in optin:
-        print(request)
+    if not len(x0):
+      # initial call with empty list
+        request, xbest, fbest = snobfit(numpy.array([]).reshape(0, len(bounds)),
+            numpy.array([]).reshape(0, 2),
+            config, dx)
+    else:
+      # initial call with just x0 as input point(s) (establishes initial request)
+        request, xbest, fbest = snobfit(x0.reshape(1, len(x0)),
+            numpy.array([objfunc(x0), math.sqrt(numpy.spacing(1))]).reshape(1, 2),
+            config, dx)
 
-    # calculate the requested points and set uncertainties
-    x, vals = fill_request(request, f, nparams)
+    if options.verbose:
+        print('request =', request)
+
+  # calculate the requested points and set uncertainties
+    x, vals = fill_request(request, objfunc, nparams)
 
     ncall0 = len(vals)                 # initial budget used
     fbestn, jbest = min_(vals[:,0])    # best function value
     xbest = x[jbest,:]
 
-    # display current number of function values, best point and function value
+  # display current number of function values, best point and function value
     log.info('# calls = %d; xbest = %s; fbest = %f', ncall0, str(xbest), fbest)
 
     nstop0 = 0;
-    # repeated calls to Snobfit
+  # repeated calls to Snobfit
     while ncall0 < budget:   # repeat till ncall function values are reached
                              # (if the stopping criterion is not fulfilled first)
         request, xbest, fbest = snobfit(x, vals, config)
-        if optin and 'verbose' in optin:
-            print(request)
+        if options.verbose:
+            print('request =', request)
 
-        # computation of the function values at the suggested points
-        x, vals = fill_request(request, f, nparams)
+      # computation of the function values at the suggested points
+        x, vals = fill_request(request, objfunc, nparams)
 
-        # update function call counter
+      # update function call counter
         ncall0 = ncall0 + len(vals)
         fbestn, jbest = min_(vals[:,0])    # best function value
         if fbestn < fbest:
             fbest = fbestn
             xbest = x[jbest,:]
 
-            # display current number of function values
+          # display current number of function values
             log.info('# calls = %d; xbest = %s; fbest = %f', ncall0, str(xbest), fbest)
 
             nstop0 = 0
         elif budget >= minfcall:
             nstop0 = nstop0 + 1
 
-        # check stopping criterion
+      # check stopping criterion
         if nstop0 >= nstop and ncall0 >= minfcall:
             break
 
-    return Result(fbest, xbest), None, None
+    return Result(fbest, xbest), objfunc.get_history()
 
 
 def snobfit(x, f, config, dx = None):
@@ -218,12 +254,13 @@ def snobfit(x, f, config, dx = None):
         f[ind,1] = math.sqrt(numpy.spacing(1))    # may be wrong
 
     rho = 0.5*(math.sqrt(5)-1)	  # golden section number
-    u1 = config['bounds'][:,0]    # lower
-    v1 = config['bounds'][:,1]    # upper
+    bounds = config['bounds']
+    u1 = bounds[:,0].reshape(1, len(bounds))    # lower
+    v1 = bounds[:,1].reshape(1, len(bounds))    # upper
 
     nreq = config['nreq']
     p = config['p']
-    n = len(u1)         # dimension of the problem
+    n = u1.shape[1]     # dimension of the problem
     nneigh = n+5        # number of nearest neighbors
 
     dy = 0.1*(v1-u1)    # defines the vector of minimal distances between two
@@ -231,22 +268,24 @@ def snobfit(x, f, config, dx = None):
 
     if dx is not None:  # a new job is started
         if numpy.any(dx<=0):
-          raise ValueError('dx should contain only positive entries')
+            raise ValueError('dx should contain only positive entries')
+
         if dx.shape[0] > 1:
-          dx = dx.T
+            dx = dx.T
+
         if x.size > 0:
-            numpy.min(numpy.app(x, [u1]))   # vertically concatenate u1 to x, grab minimum row
-            numpy.max(numpy.app(x, [v1]))   # vertically concatenate v1 to x, grab maximum row
+            u = numpy.minimum(x.min(axis=0), u1)
+            v = numpy.maximum(x.max(axis=0), v1)
         else:
-            u = u1
-            v = v1
+            u = u1.copy()
+            v = v1.copy()
 
         x, f, np, t = snobinput(x, f)   # throw out duplicates among the points
                                         # and compute mean function value and
                                         # deviation
         if x.size > 0:
             xl, xu, x, f, nsplit, small = snobsplit(x, f, u, v, None, u, v)
-            d = inf * ones((1, len(x)))
+            d = numpy.inf*numpy.ones((1, len(x)))
         else:
             xl = numpy.array([])
             xu = numpy.array([])
@@ -261,16 +300,16 @@ def snobfit(x, f, config, dx = None):
             fmn = 1
             fmx = 0
 
-        if (len(x) >= nneigh+1) and (fmn < fmx):
+        if len(x) >= nneigh+1 and fmn < fmx:
             inew = range(len(x))
             near = numpy.zeros((len(x), nneigh))
             d = numpy.zeros(len(x))
             for j in inew:
-                near[j], d[j] = snobnn(x[j],x,nneigh,dx)
+                near[j], d[j] = snobnn(x[j], x, nneigh, dx)
 
             fnan = find(numpy.isnan(f[:,0]))
             if fnan.size > 0:
-                f = snobnan(fnan,f,near,inew)
+                f = snobnan(fnan, f, near, inew)
 
             jsize = inew[-1]
             y = numpy.zeros((jsize, 2))
@@ -288,10 +327,10 @@ def snobfit(x, f, config, dx = None):
             near = numpy.array([], dtype=int)
             d = numpy.inf*numpy.ones((1,len(x)))
 
-            x1 = snob5(x, u1, v1, dx, nreq)
-            request = numpy.concatenate((x1, numpy.nan*numpy.ones((nreq,1)), 5*numpy.ones((nreq,1))), 1)
-            if x.size > 0:
-                (fbest, jbest) = min_(f[:,1])
+            x5 = snob5(x, u1, v1, dx, nreq)
+            request = numpy.concatenate((x5, numpy.nan*numpy.ones((nreq,1)), 5*numpy.ones((nreq,1))), 1)
+            if x.size > 0 and f.size > 0:
+                fbest, jbest = min_(f[:,0])
                 xbest = x[jbest]
             else:
                 xbest = numpy.nan*numpy.ones((1,n))
@@ -307,6 +346,7 @@ def snobfit(x, f, config, dx = None):
         xnew = x.copy()
         fnew = f.copy()
         xbest, fbest, x, f, xl, xu, y, nsplit, small, near, d, np, t, fnan, u, v, dx = _snobload()
+
         nx = len(xnew)
         oldxbest = xbest
 
@@ -329,14 +369,16 @@ def snobfit(x, f, config, dx = None):
                 y = numpy.append(y, numpy.zeros((jsize-len(y), x.shape[1])), axis=0)
             g = numpy.zeros((jsize, x.shape[1]))
             sigma = numpy.zeros(jsize)
-            f = extend(f, x.shape[1]-1)
+            if f.shape[1] < 3:
+               f = extend(f, 3-f.shape[1])
+
             for j in inew:
                 y[j], f[j,2], c, sigma[j] = snoblocf(j, x, f[:,0:2], near, dx, u, v)
                 g[j] = c
 
         else:
-            x1 = snob5(x, u1, v1, dx, nreq)
-            request = numpy.concatenate((x1, numpy.NaN*numpy.ones((len(x1),1)), 5*numpy.ones((len(x1),1))), 1)
+            x5 = snob5(x, u1, v1, dx, nreq)
+            request = numpy.concatenate((x5, numpy.NaN*numpy.ones((nreq,1)), 5*numpy.ones((nreq,1))), 1)
             if x.size > 0:
                 (fbest, ibest) = min_(f[:,0])
                 xbest = x[ibest]
@@ -383,7 +425,7 @@ def snobfit(x, f, config, dx = None):
         if dmin <= 0.05*dmax:
             isplit = numpy.append(isplit, j)
         else:
-            request = numpy.vstack((request, numpy.concatenate((z, numpy.array((f1, 1))), 0)))
+            request = numpy.vstack((request, numpy.concatenate((z, numpy.array((f1, 1), ndmin=2)), axis=1)))
 
     if len(request) < nreq:
         globloc = nreq - len(request)
@@ -423,11 +465,11 @@ def snobfit(x, f, config, dx = None):
                     if numpy.sum(y1 == y[l0]) < n:
                         D = f[l0,1]/dx**2
                         #Possibly problem area!
-                        f1 = f[l0,0] + g[l0].dot((y1 - x[l0]).T) + sigma[l0].dot( \
-                            (y1 - x[l0]).dot(diag(D).dot((y1-x(l0)).T) + f[l0,1]))
+                        f1 = f[l0,0] + g[l0].dot((y1 - x[l0]).T) + sigma[l0]*( \
+                            (y1 - x[l0]).dot(diag(D).dot((y1-x[l0]).T) + f[l0,1]))
                     else:
                         f1 = f[l0,2]
-                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 2))), 0)))
+                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 2), ndmin=2)), axis=1)))
 
                 sreq = len(request)
                 j += 1
@@ -459,11 +501,11 @@ def snobfit(x, f, config, dx = None):
                     if numpy.sum(y1==y[l0]) < n:
                         D = f[l0,1]/(dx**2)
                         f1 = f[l0,0] + g[l0].dot((y1 - x[l0]).T) + \
-                        sigma[l0].dot(((y1 - x[l0]).dot(diag(D).dot((y1 - x[l0]).T))) + f[l0,1])
+                        sigma[l0]*(((y1 - x[l0]).dot(diag(D).dot((y1 - x[l0]).T))) + f[l0,1])
                     else:
                         f1 = f[l0,2]
 
-                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 3))), 0)))
+                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 3), ndmin=2)), axis=1)))
 
                 sreq = len(request)
                 j += 1
@@ -479,7 +521,7 @@ def snobfit(x, f, config, dx = None):
                 (not sreq or numpy.min( \
                      numpy.max(numpy.abs(request[:,:n] - numpy.outer(numpy.ones(sreq), y1)) - \
                      numpy.outer(numpy.ones(sreq), dx), axis=1)) >= -numpy.spacing(1)):
-            request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 4))), 0)))
+            request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 4), ndmin=2)), axis=1)))
 
         sreq = len(request)
         if sreq == nreq:
@@ -487,7 +529,7 @@ def snobfit(x, f, config, dx = None):
 
     first = True
     while (sreq < nreq) and ind.size > 0:   # and find(small[ind] <= (minsmall + m1)).any():
-        for m in range(m1+1):
+        for m in range(int(m1+1)):
             if first:
                 first = False
                 continue
@@ -510,26 +552,28 @@ def snobfit(x, f, config, dx = None):
                         (not sreq or numpy.min( \
                              numpy.max(numpy.abs(request[:,:n] - numpy.outer(numpy.ones(sreq), y1)) - \
                              numpy.outer(numpy.ones(sreq), numpy.maximum(dy,dx)), axis=1)) >= -numpy.spacing(1)):
-                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 4))), 0)))
+                    request = numpy.vstack((request, numpy.concatenate((y1, numpy.array((f1, 4), ndmin=2)), axis=1)))
 
                 sreq = len(request)
                 if sreq == nreq:
                     break
             m = 0
 
-    if len(request.shape) < nreq:
-        x1 = snob5(numpy.concatenate((x, request[:,:n])), u1, v1, dx, nreq - len(request))
+    if len(request) < nreq:
+        x5 = snob5(numpy.concatenate((x, request[:,:n])), u1, v1, dx, nreq - len(request))
         nx = len(x)
-        for j in range(len(x1)):
-            i = find((numpy.sum(xl <= numpy.outer(numpy.ones(nx), x1[j])) and \
-                   (numpy.outer(numpy.ones((nx,1)), x1[j]) <= xu), 1) == n)
+        for j in range(len(x5)):
+            x5j = x5[(j,),:]
+            i = find(numpy.sum(numpy.logical_and( \
+                         xl <= numpy.outer(numpy.ones(nx), x5j), numpy.outer(numpy.ones((nx,1)), x5j) <= xu),
+                     1) == n)
             if len(i) > 1:
                 minv, i1 = min_(small[i])
                 i = i[i1]
 
             D = f[i,1]/(dx**2)
-            f1 = f[i,0] + (x1[j] - x[i]).dot(g[i].T) + sigma[i].dot((x1[j] - x[i]).dot(diag(D).dot((x1[j] - x[i]).T))) + f[i,1]
-            request = numpy.vstack((request, numpy.concatenate((x1[j], numpy.array((f1, 5))), 0)))
+            f1 = f[i,0] + (x5j - x[i]).dot(g[i].T) + sigma[i]*((x5j - x[i]).dot(diag(D).dot((x5j - x[i]).T))) + f[i,1]
+            request = numpy.vstack((request, numpy.concatenate((x5j, numpy.array((f1, 5), ndmin=2)), axis=1)))
 
     if len(request) < nreq:
         snobwarn()
